from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncIterator

from pydantic import Field

from unstructured_ingest.error import (
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    FileData,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.v2.processes.connectors.onedrive import (
    OnedriveAccessConfig,
    OnedriveConnectionConfig,
    OnedriveDownloader,
    OnedriveDownloaderConfig,
    OnedriveIndexer,
    OnedriveIndexerConfig,
)

if TYPE_CHECKING:
    from office365.onedrive.driveitems.driveItem import DriveItem

CONNECTOR_TYPE = "sharepoint"
LEGACY_DEFAULT_PATH = "Shared Documents"


class SharepointAccessConfig(OnedriveAccessConfig):
    client_cred: str = Field(description="Microsoft App client secret")


class SharepointConnectionConfig(OnedriveConnectionConfig):
    site: str = Field(
        description="Sharepoint site url. Process either base url e.g \
                    https://[tenant].sharepoint.com  or relative sites \
                    https://[tenant].sharepoint.com/sites/<site_name>. \
                    To process all sites within the tenant pass a site url as \
                    https://[tenant]-admin.sharepoint.com.\
                    This requires the app to be registered at a tenant level"
    )


class SharepointIndexerConfig(OnedriveIndexerConfig):
    pass


@dataclass
class SharepointIndexer(OnedriveIndexer):
    connection_config: SharepointConnectionConfig
    index_config: SharepointIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["office365"], extras="sharepoint")
    async def run_async(self, **kwargs: Any) -> AsyncIterator[FileData]:
        from office365.runtime.client_request_exception import ClientRequestException

        token_resp = await asyncio.to_thread(self.connection_config.get_token)
        if "error" in token_resp:
            raise SourceConnectionError(
                f"[{self.connector_type}]: {token_resp['error']} "
                f"({token_resp.get('error_description')})"
            )

        client = await asyncio.to_thread(self.connection_config.get_client)
        try:
            site = client.sites.get_by_url(self.connection_config.site).get().execute_query()
            site_drive_item = site.drive.get().execute_query().root
        except ClientRequestException:
            logger.info("Site not found")

        path = self.index_config.path
        # Deprecated sharepoint sdk needed a default path. Microsoft Graph SDK does not.
        if path and path != LEGACY_DEFAULT_PATH:
            site_drive_item = site_drive_item.get_by_path(path).get().execute_query()

        for drive_item in site_drive_item.get_files(
            recursive=self.index_config.recursive
        ).execute_query():
            file_data = await self.drive_item_to_file_data(drive_item=drive_item)
            yield file_data


class SharepointDownloaderConfig(OnedriveDownloaderConfig):
    pass


@dataclass
class SharepointDownloader(OnedriveDownloader):
    connection_config: SharepointConnectionConfig
    download_config: SharepointDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["office365"], extras="onedrive")
    def _fetch_file(self, file_data: FileData) -> DriveItem:
        from office365.runtime.client_request_exception import ClientRequestException

        if file_data.source_identifiers is None or not file_data.source_identifiers.fullpath:
            raise ValueError(
                f"file data doesn't have enough information to get "
                f"file content: {file_data.model_dump()}"
            )

        server_relative_path = file_data.source_identifiers.fullpath
        client = self.connection_config.get_client()

        try:
            site = client.sites.get_by_url(self.connection_config.site).get().execute_query()
            site_drive_item = site.drive.get().execute_query().root
        except ClientRequestException:
            logger.info("Site not found")
        file = site_drive_item.get_by_path(server_relative_path).get().execute_query()

        if not file:
            raise FileNotFoundError(f"file not found: {server_relative_path}")
        return file


sharepoint_source_entry = SourceRegistryEntry(
    connection_config=SharepointConnectionConfig,
    indexer_config=SharepointIndexerConfig,
    indexer=SharepointIndexer,
    downloader_config=SharepointDownloaderConfig,
    downloader=SharepointDownloader,
)



from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Generator, List, Optional

from pydantic import Field, Secret

from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.html import HtmlMixin
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    FileData,
    FileDataSourceMetadata,
    Indexer,
    IndexerConfig,
    SourceIdentifiers,
    download_responses,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)

if TYPE_CHECKING:
    from atlassian import Confluence

CONNECTOR_TYPE = "confluence"


class ConfluenceAccessConfig(AccessConfig):
    password: Optional[str] = Field(
        description="Confluence password or Cloud API token",
        default=None,
    )
    token: Optional[str] = Field(
        description="Confluence Personal Access Token",
        default=None,
    )


class ConfluenceConnectionConfig(ConnectionConfig):
    url: str = Field(description="URL of the Confluence instance")
    username: Optional[str] = Field(
        description="Username or email for authentication",
        default=None,
    )
    cloud: bool = Field(description="Authenticate to Confluence Cloud", default=False)
    access_config: Secret[ConfluenceAccessConfig] = Field(
        description="Access configuration for Confluence"
    )

    def model_post_init(self, __context):
        access_configs = self.access_config.get_secret_value()
        basic_auth = self.username and access_configs.password
        pat_auth = access_configs.token
        if self.cloud and not basic_auth:
            raise ValueError(
                "cloud authentication requires username and API token (--password), "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if basic_auth and pat_auth:
            raise ValueError(
                "both password and token provided, only one allowed, "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if not (basic_auth or pat_auth):
            raise ValueError(
                "no form of auth provided, see: https://atlassian-python-api.readthedocs.io/"
            )

    @requires_dependencies(["atlassian"], extras="confluence")
    @contextmanager
    def get_client(self) -> "Confluence":
        from atlassian import Confluence

        access_configs = self.access_config.get_secret_value()
        with Confluence(
            url=self.url,
            username=self.username,
            password=access_configs.password,
            token=access_configs.token,
            cloud=self.cloud,
        ) as client:
            yield client


class ConfluenceIndexerConfig(IndexerConfig):
    max_num_of_spaces: int = Field(500, description="Maximum number of spaces to index")
    max_num_of_docs_from_each_space: int = Field(
        100, description="Maximum number of documents to fetch from each space"
    )
    spaces: Optional[List[str]] = Field(None, description="List of specific space keys to index")


@dataclass
class ConfluenceIndexer(Indexer):
    connection_config: ConfluenceConnectionConfig
    index_config: ConfluenceIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> bool:
        try:

            # Attempt to retrieve a list of spaces with limit=1.
            # This should only succeed if all creds are valid
            with self.connection_config.get_client() as client:
                client.get_all_spaces(limit=1)
            logger.info("Connection to Confluence successful.")
            return True
        except Exception as e:
            logger.error(f"Failed to connect to Confluence: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to connect to Confluence: {e}")

    def _get_space_ids(self) -> List[str]:
        spaces = self.index_config.spaces
        if spaces:
            return spaces
        else:
            with self.connection_config.get_client() as client:
                all_spaces = client.get_all_spaces(limit=self.index_config.max_num_of_spaces)
            space_ids = [space["key"] for space in all_spaces["results"]]
            return space_ids

    def _get_docs_ids_within_one_space(self, space_id: str) -> List[dict]:
        with self.connection_config.get_client() as client:
            pages = client.get_all_pages_from_space(
                space=space_id,
                start=0,
                limit=self.index_config.max_num_of_docs_from_each_space,
                expand=None,
                content_type="page",
                status=None,
            )
        doc_ids = [{"space_id": space_id, "doc_id": page["id"]} for page in pages]
        return doc_ids

    def run(self) -> Generator[FileData, None, None]:
        from time import time

        space_ids = self._get_space_ids()
        for space_id in space_ids:
            doc_ids = self._get_docs_ids_within_one_space(space_id)
            for doc in doc_ids:
                doc_id = doc["doc_id"]
                # Build metadata
                metadata = FileDataSourceMetadata(
                    date_processed=str(time()),
                    url=f"{self.connection_config.url}/pages/{doc_id}",
                    record_locator={
                        "space_id": space_id,
                        "document_id": doc_id,
                    },
                )
                additional_metadata = {
                    "space_id": space_id,
                    "document_id": doc_id,
                }

                # Construct relative path and filename
                filename = f"{doc_id}.html"
                relative_path = str(Path(space_id) / filename)

                source_identifiers = SourceIdentifiers(
                    filename=filename,
                    fullpath=relative_path,
                    rel_path=relative_path,
                )

                file_data = FileData(
                    identifier=doc_id,
                    connector_type=self.connector_type,
                    metadata=metadata,
                    additional_metadata=additional_metadata,
                    source_identifiers=source_identifiers,
                )
                yield file_data


class ConfluenceDownloaderConfig(DownloaderConfig, HtmlMixin):
    pass


@dataclass
class ConfluenceDownloader(Downloader):
    connection_config: ConfluenceConnectionConfig
    download_config: ConfluenceDownloaderConfig = field(default_factory=ConfluenceDownloaderConfig)
    connector_type: str = CONNECTOR_TYPE

    def download_embedded_files(
        self, session, html: str, current_file_data: FileData
    ) -> list[DownloadResponse]:
        if not self.download_config.extract_files:
            return []
        url = current_file_data.metadata.url
        if url is None:
            logger.warning(
                f"""Missing URL for file: {current_file_data.source_identifiers.filename}.
                Skipping file extraction."""
            )
            return []
        filepath = current_file_data.source_identifiers.relative_path
        download_path = Path(self.download_dir) / filepath
        download_dir = download_path.with_suffix("")
        return self.download_config.extract_embedded_files(
            url=url,
            download_dir=download_dir,
            original_filedata=current_file_data,
            html=html,
            session=session,
        )

    def run(self, file_data: FileData, **kwargs) -> download_responses:
        from bs4 import BeautifulSoup

        doc_id = file_data.identifier
        try:
            with self.connection_config.get_client() as client:
                page = client.get_page_by_id(
                    page_id=doc_id,
                    expand="history.lastUpdated,version,body.view",
                )

        except Exception as e:
            logger.error(f"Failed to retrieve page with ID {doc_id}: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to retrieve page with ID {doc_id}: {e}")

        if not page:
            raise ValueError(f"Page with ID {doc_id} does not exist.")

        content = page["body"]["view"]["value"]
        # This supports v2 html parsing in unstructured
        title = page["title"]
        title_html = f"<title>{title}</title>"
        content = f"<body class='Document' >{title_html}{content}</body>"
        if self.download_config.extract_images:
            with self.connection_config.get_client() as client:
                content = self.download_config.extract_html_images(
                    url=file_data.metadata.url, html=content, session=client._session
                )

        filepath = file_data.source_identifiers.relative_path
        download_path = Path(self.download_dir) / filepath
        download_path.parent.mkdir(parents=True, exist_ok=True)
        with open(download_path, "w", encoding="utf8") as f:
            soup = BeautifulSoup(content, "html.parser")
            f.write(soup.prettify())

        # Update file_data with metadata
        file_data.metadata.date_created = page["history"]["createdDate"]
        file_data.metadata.date_modified = page["version"]["when"]
        file_data.metadata.version = str(page["version"]["number"])
        file_data.display_name = title

        download_response = self.generate_download_response(
            file_data=file_data, download_path=download_path
        )
        if self.download_config.extract_files:
            with self.connection_config.get_client() as client:
                extracted_download_responses = self.download_embedded_files(
                    html=content,
                    current_file_data=download_response["file_data"],
                    session=client._session,
                )
                if extracted_download_responses:
                    for dr in extracted_download_responses:
                        fd = dr["file_data"]
                        source_file_path = Path(file_data.source_identifiers.fullpath).with_suffix(
                            ""
                        )
                        new_fullpath = source_file_path / fd.source_identifiers.filename
                        fd.source_identifiers = SourceIdentifiers(
                            fullpath=new_fullpath.as_posix(), filename=new_fullpath.name
                        )
                    extracted_download_responses.append(download_response)
                    return extracted_download_responses
        return download_response


confluence_source_entry = SourceRegistryEntry(
    connection_config=ConfluenceConnectionConfig,
    indexer_config=ConfluenceIndexerConfig,
    indexer=ConfluenceIndexer,
    downloader_config=ConfluenceDownloaderConfig,
    downloader=ConfluenceDownloader,
)



import json
import os
import time
from pathlib import Path
from uuid import uuid4

import pytest
from _pytest.fixtures import TopRequest
from azure.core.credentials import AzureKeyCredential
from azure.search.documents import SearchClient
from azure.search.documents.indexes import SearchIndexClient
from azure.search.documents.indexes.models import (
    ComplexField,
    CorsOptions,
    HnswAlgorithmConfiguration,
    HnswParameters,
    SearchField,
    SearchFieldDataType,
    SearchIndex,
    SimpleField,
    VectorSearch,
    VectorSearchAlgorithmMetric,
    VectorSearchProfile,
)

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.azure_ai_search import (
    CONNECTOR_TYPE,
    RECORD_ID_LABEL,
    AzureAISearchAccessConfig,
    AzureAISearchConnectionConfig,
    AzureAISearchUploader,
    AzureAISearchUploaderConfig,
    AzureAISearchUploadStager,
    AzureAISearchUploadStagerConfig,
)

repo_path = Path(__file__).parent.resolve()

API_KEY = "AZURE_SEARCH_API_KEY"
ENDPOINT = "https://ingest-test-azure-cognitive-search.search.windows.net"


def get_api_key() -> str:
    key = os.environ[API_KEY]
    return key


def get_fields() -> list:
    data_source_fields = [
        SimpleField(name="url", type=SearchFieldDataType.String),
        SimpleField(name="version", type=SearchFieldDataType.String),
        SimpleField(name="date_created", type=SearchFieldDataType.DateTimeOffset),
        SimpleField(name="date_modified", type=SearchFieldDataType.DateTimeOffset),
        SimpleField(name="date_processed", type=SearchFieldDataType.DateTimeOffset),
        SimpleField(name="permissions_data", type=SearchFieldDataType.String),
        SimpleField(name="record_locator", type=SearchFieldDataType.String),
    ]
    coordinates_fields = [
        SimpleField(name="system", type=SearchFieldDataType.String),
        SimpleField(name="layout_width", type=SearchFieldDataType.Double),
        SimpleField(name="layout_height", type=SearchFieldDataType.Double),
        SimpleField(name="points", type=SearchFieldDataType.String),
    ]
    metadata_fields = [
        SimpleField(name="orig_elements", type=SearchFieldDataType.String),
        SimpleField(name="category_depth", type=SearchFieldDataType.Int32),
        SimpleField(name="parent_id", type=SearchFieldDataType.String),
        SimpleField(name="attached_to_filename", type=SearchFieldDataType.String),
        SimpleField(name="filetype", type=SearchFieldDataType.String),
        SimpleField(name="last_modified", type=SearchFieldDataType.DateTimeOffset),
        SimpleField(name="is_continuation", type=SearchFieldDataType.Boolean),
        SimpleField(name="file_directory", type=SearchFieldDataType.String),
        SimpleField(name="filename", type=SearchFieldDataType.String),
        ComplexField(name="data_source", fields=data_source_fields),
        ComplexField(name="coordinates", fields=coordinates_fields),
        SimpleField(
            name="languages", type=SearchFieldDataType.Collection(SearchFieldDataType.String)
        ),
        SimpleField(name="page_number", type=SearchFieldDataType.String),
        SimpleField(name="links", type=SearchFieldDataType.Collection(SearchFieldDataType.String)),
        SimpleField(name="page_name", type=SearchFieldDataType.String),
        SimpleField(name="url", type=SearchFieldDataType.String),
        SimpleField(
            name="link_urls", type=SearchFieldDataType.Collection(SearchFieldDataType.String)
        ),
        SimpleField(
            name="link_texts", type=SearchFieldDataType.Collection(SearchFieldDataType.String)
        ),
        SimpleField(
            name="sent_from", type=SearchFieldDataType.Collection(SearchFieldDataType.String)
        ),
        SimpleField(
            name="sent_to", type=SearchFieldDataType.Collection(SearchFieldDataType.String)
        ),
        SimpleField(name="subject", type=SearchFieldDataType.String),
        SimpleField(name="section", type=SearchFieldDataType.String),
        SimpleField(name="header_footer_type", type=SearchFieldDataType.String),
        SimpleField(
            name="emphasized_text_contents",
            type=SearchFieldDataType.Collection(SearchFieldDataType.String),
        ),
        SimpleField(
            name="emphasized_text_tags",
            type=SearchFieldDataType.Collection(SearchFieldDataType.String),
        ),
        SimpleField(name="text_as_html", type=SearchFieldDataType.String),
        SimpleField(name="regex_metadata", type=SearchFieldDataType.String),
        SimpleField(name="detection_class_prob", type=SearchFieldDataType.Double),
    ]
    fields = [
        SimpleField(name="id", type=SearchFieldDataType.String, key=True),
        SimpleField(name=RECORD_ID_LABEL, type=SearchFieldDataType.String, filterable=True),
        SimpleField(name="element_id", type=SearchFieldDataType.String),
        SimpleField(name="text", type=SearchFieldDataType.String),
        SimpleField(name="type", type=SearchFieldDataType.String),
        ComplexField(name="metadata", fields=metadata_fields),
        SearchField(
            name="embeddings",
            type=SearchFieldDataType.Collection(SearchFieldDataType.Single),
            vector_search_dimensions=384,
            vector_search_profile_name="embeddings-config-profile",
        ),
    ]
    return fields


def get_vector_search() -> VectorSearch:
    return VectorSearch(
        algorithms=[
            HnswAlgorithmConfiguration(
                name="hnsw-config",
                parameters=HnswParameters(
                    metric=VectorSearchAlgorithmMetric.COSINE,
                ),
            )
        ],
        profiles=[
            VectorSearchProfile(
                name="embeddings-config-profile", algorithm_configuration_name="hnsw-config"
            )
        ],
    )


def get_search_index_client() -> SearchIndexClient:
    api_key = get_api_key()
    return SearchIndexClient(ENDPOINT, AzureKeyCredential(api_key))


@pytest.fixture
def index() -> str:
    random_id = str(uuid4())[:8]
    index_name = f"utic-test-{random_id}"
    client = get_search_index_client()
    index = SearchIndex(
        name=index_name,
        fields=get_fields(),
        vector_search=get_vector_search(),
        cors_options=CorsOptions(allowed_origins=["*"], max_age_in_seconds=60),
    )
    print(f"creating index: {index_name}")
    client.create_index(index=index)
    try:
        yield index_name
    finally:
        print(f"deleting index: {index_name}")
        client.delete_index(index)


def validate_count(
    search_client: SearchClient, expected_count: int, retries: int = 10, interval: int = 1
) -> None:
    index_count = search_client.get_document_count()
    if index_count == expected_count:
        return
    tries = 0
    while tries < retries:
        time.sleep(interval)
        index_count = search_client.get_document_count()
        if index_count == expected_count:
            break
    assert index_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from index: {index_count}"
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@requires_env("AZURE_SEARCH_API_KEY")
async def test_azure_ai_search_destination(
    upload_file: Path,
    index: str,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    stager = AzureAISearchUploadStager(upload_stager_config=AzureAISearchUploadStagerConfig())

    uploader = AzureAISearchUploader(
        connection_config=AzureAISearchConnectionConfig(
            access_config=AzureAISearchAccessConfig(key=get_api_key()),
            endpoint=ENDPOINT,
            index=index,
        ),
        upload_config=AzureAISearchUploaderConfig(),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    with uploader.connection_config.get_search_client() as search_client:
        validate_count(search_client=search_client, expected_count=expected_count)

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    with uploader.connection_config.get_search_client() as search_client:
        validate_count(search_client=search_client, expected_count=expected_count)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_azure_ai_search_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = AzureAISearchUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import os

import pytest

from test.integration.connectors.utils.constants import SOURCE_TAG, UNCATEGORIZED_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.confluence import (
    CONNECTOR_TYPE,
    ConfluenceAccessConfig,
    ConfluenceConnectionConfig,
    ConfluenceDownloader,
    ConfluenceDownloaderConfig,
    ConfluenceIndexer,
    ConfluenceIndexerConfig,
)


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("CONFLUENCE_USER_EMAIL", "CONFLUENCE_API_TOKEN")
async def test_confluence_source(temp_dir):
    # Retrieve environment variables
    confluence_url = "https://unstructured-ingest-test.atlassian.net"
    user_email = os.environ["CONFLUENCE_USER_EMAIL"]
    api_token = os.environ["CONFLUENCE_API_TOKEN"]
    spaces = ["testteamsp", "MFS"]

    # Create connection and indexer configurations
    access_config = ConfluenceAccessConfig(password=api_token)
    connection_config = ConfluenceConnectionConfig(
        url=confluence_url,
        username=user_email,
        access_config=access_config,
    )
    index_config = ConfluenceIndexerConfig(
        max_num_of_spaces=500,
        max_num_of_docs_from_each_space=100,
        spaces=spaces,
    )

    download_config = ConfluenceDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = ConfluenceIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = ConfluenceDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="confluence",
            expected_num_files=11,
            validate_downloaded_files=True,
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("CONFLUENCE_USER_EMAIL", "CONFLUENCE_API_TOKEN")
async def test_confluence_source_large(temp_dir):
    # Retrieve environment variables
    confluence_url = "https://unstructured-ingest-test.atlassian.net"
    user_email = os.environ["CONFLUENCE_USER_EMAIL"]
    api_token = os.environ["CONFLUENCE_API_TOKEN"]
    spaces = ["testteamsp1"]

    # Create connection and indexer configurations
    access_config = ConfluenceAccessConfig(password=api_token)
    connection_config = ConfluenceConnectionConfig(
        url=confluence_url,
        username=user_email,
        access_config=access_config,
    )
    index_config = ConfluenceIndexerConfig(
        max_num_of_spaces=10,
        max_num_of_docs_from_each_space=250,
        spaces=spaces,
    )

    download_config = ConfluenceDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = ConfluenceIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = ConfluenceDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="confluence_large", expected_num_files=250, validate_file_data=False
        ),
    )



import os
from pathlib import Path
from typing import Literal, Union
from uuid import uuid4

import lancedb
import pandas as pd
import pyarrow as pa
import pytest
import pytest_asyncio
from lancedb import AsyncConnection
from upath import UPath

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.interfaces.file_data import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.lancedb.aws import (
    LanceDBAwsAccessConfig,
    LanceDBAwsConnectionConfig,
    LanceDBAwsUploader,
)
from unstructured_ingest.v2.processes.connectors.lancedb.azure import (
    LanceDBAzureAccessConfig,
    LanceDBAzureConnectionConfig,
    LanceDBAzureUploader,
)
from unstructured_ingest.v2.processes.connectors.lancedb.gcp import (
    LanceDBGCSAccessConfig,
    LanceDBGCSConnectionConfig,
    LanceDBGSPUploader,
)
from unstructured_ingest.v2.processes.connectors.lancedb.lancedb import (
    CONNECTOR_TYPE,
    LanceDBUploaderConfig,
    LanceDBUploadStager,
)
from unstructured_ingest.v2.processes.connectors.lancedb.local import (
    LanceDBLocalAccessConfig,
    LanceDBLocalConnectionConfig,
    LanceDBLocalUploader,
)

DATABASE_NAME = "database"
TABLE_NAME = "elements"
DIMENSION = 384
NUMBER_EXPECTED_ROWS = 22
S3_BUCKET = "s3://utic-ingest-test-fixtures/"
GS_BUCKET = "gs://utic-test-ingest-fixtures-output/"
AZURE_BUCKET = "az://utic-ingest-test-fixtures-output/"
REQUIRED_ENV_VARS = {
    "s3": ("S3_INGEST_TEST_ACCESS_KEY", "S3_INGEST_TEST_SECRET_KEY"),
    "gcs": ("GCP_INGEST_SERVICE_KEY",),
    "az": ("AZURE_DEST_CONNECTION_STR",),
    "local": (),
}

SCHEMA = pa.schema(
    [
        pa.field(RECORD_ID_LABEL, pa.string()),
        pa.field("vector", pa.list_(pa.float16(), DIMENSION)),
        pa.field("text", pa.string(), nullable=True),
        pa.field("type", pa.string(), nullable=True),
        pa.field("element_id", pa.string(), nullable=True),
        pa.field("metadata-text_as_html", pa.string(), nullable=True),
        pa.field("metadata-filetype", pa.string(), nullable=True),
        pa.field("metadata-filename", pa.string(), nullable=True),
        pa.field("metadata-languages", pa.list_(pa.string()), nullable=True),
        pa.field("metadata-is_continuation", pa.bool_(), nullable=True),
        pa.field("metadata-page_number", pa.int32(), nullable=True),
    ]
)
NUMBER_EXPECTED_COLUMNS = len(SCHEMA.names)


@pytest_asyncio.fixture
async def connection_with_uri(request, tmp_path: Path):
    target = request.param
    uri = _get_uri(target, local_base_path=tmp_path)

    unset_variables = [env for env in REQUIRED_ENV_VARS[target] if env not in os.environ]
    if unset_variables:
        pytest.skip(
            reason="Following required environment variables were not set: "
            + f"{', '.join(unset_variables)}"
        )

    storage_options = {
        "aws_access_key_id": os.getenv("S3_INGEST_TEST_ACCESS_KEY"),
        "aws_secret_access_key": os.getenv("S3_INGEST_TEST_SECRET_KEY"),
        "google_service_account_key": os.getenv("GCP_INGEST_SERVICE_KEY"),
    }
    azure_connection_string = os.getenv("AZURE_DEST_CONNECTION_STR")
    if azure_connection_string:
        storage_options.update(_parse_azure_connection_string(azure_connection_string))

    storage_options = {key: value for key, value in storage_options.items() if value is not None}
    connection = await lancedb.connect_async(
        uri=uri,
        storage_options=storage_options,
    )
    await connection.create_table(name=TABLE_NAME, schema=SCHEMA)

    yield connection, uri

    await connection.drop_database()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.parametrize("connection_with_uri", ["local", "s3", "gcs", "az"], indirect=True)
async def test_lancedb_destination(
    upload_file: Path,
    connection_with_uri: tuple[AsyncConnection, str],
    tmp_path: Path,
) -> None:
    connection, uri = connection_with_uri
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock-file-data",
    )
    stager = LanceDBUploadStager()
    uploader = _get_uploader(uri)
    staged_file_path = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )

    await uploader.run_async(path=staged_file_path, file_data=file_data)

    # Test upload to empty table
    with await connection.open_table(TABLE_NAME) as table:
        table_df: pd.DataFrame = await table.to_pandas()

    assert len(table_df) == NUMBER_EXPECTED_ROWS
    assert len(table_df.columns) == NUMBER_EXPECTED_COLUMNS

    assert table_df[RECORD_ID_LABEL][0] == file_data.identifier
    assert table_df["element_id"][0] == "2470d8dc42215b3d68413b55bf00fed2"
    assert table_df["type"][0] == "CompositeElement"
    assert table_df["metadata-filename"][0] == "DA-1p-with-duplicate-pages.pdf.json"
    assert table_df["metadata-text_as_html"][0] is None

    # Test upload of the second file, rows should be appended
    file_data.identifier = "mock-file-data-2"
    staged_second_file_path = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=f"{upload_file.stem}-2{upload_file.suffix}",
    )
    await uploader.run_async(path=staged_second_file_path, file_data=file_data)
    with await connection.open_table(TABLE_NAME) as table:
        appended_table_df: pd.DataFrame = await table.to_pandas()
    assert len(appended_table_df) == 2 * NUMBER_EXPECTED_ROWS

    # Test re-upload of the first file, rows should be overwritten, not appended
    await uploader.run_async(path=staged_file_path, file_data=file_data)
    with await connection.open_table(TABLE_NAME) as table:
        overwritten_table_df: pd.DataFrame = await table.to_pandas()
    assert len(overwritten_table_df) == 2 * NUMBER_EXPECTED_ROWS


class TestPrecheck:
    @pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
    @pytest.mark.parametrize("connection_with_uri", ["local", "s3", "gcs", "az"], indirect=True)
    def test_succeeds(
        self,
        upload_file: Path,
        connection_with_uri: tuple[AsyncConnection, str],
        tmp_path: Path,
    ) -> None:
        _, uri = connection_with_uri
        uploader = _get_uploader(uri)
        uploader.precheck()


def _get_uri(target: Literal["local", "s3", "gcs", "az"], local_base_path: Path) -> str:
    if target == "local":
        return str(local_base_path / DATABASE_NAME)
    if target == "s3":
        base_uri = UPath(S3_BUCKET)
    elif target == "gcs":
        base_uri = UPath(GS_BUCKET)
    elif target == "az":
        base_uri = UPath(AZURE_BUCKET)

    return str(base_uri / "destination" / "lancedb" / str(uuid4()) / DATABASE_NAME)


def _get_uploader(
    uri: str,
) -> Union[LanceDBAzureUploader, LanceDBAzureUploader, LanceDBAwsUploader, LanceDBGSPUploader]:
    target = uri.split("://", maxsplit=1)[0] if uri.startswith(("s3", "az", "gs")) else "local"
    upload_config = LanceDBUploaderConfig(table_name=TABLE_NAME)
    if target == "az":
        azure_connection_string = os.getenv("AZURE_DEST_CONNECTION_STR")
        access_config_kwargs = _parse_azure_connection_string(azure_connection_string)
        return LanceDBAzureUploader(
            upload_config=upload_config,
            connection_config=LanceDBAzureConnectionConfig(
                access_config=LanceDBAzureAccessConfig(**access_config_kwargs),
                uri=uri,
            ),
        )

    elif target == "s3":
        return LanceDBAwsUploader(
            upload_config=upload_config,
            connection_config=LanceDBAwsConnectionConfig(
                access_config=LanceDBAwsAccessConfig(
                    aws_access_key_id=os.getenv("S3_INGEST_TEST_ACCESS_KEY"),
                    aws_secret_access_key=os.getenv("S3_INGEST_TEST_SECRET_KEY"),
                ),
                uri=uri,
            ),
        )
    elif target == "gs":
        return LanceDBGSPUploader(
            upload_config=upload_config,
            connection_config=LanceDBGCSConnectionConfig(
                access_config=LanceDBGCSAccessConfig(
                    google_service_account_key=os.getenv("GCP_INGEST_SERVICE_KEY")
                ),
                uri=uri,
            ),
        )
    else:
        return LanceDBLocalUploader(
            upload_config=upload_config,
            connection_config=LanceDBLocalConnectionConfig(
                access_config=LanceDBLocalAccessConfig(),
                uri=uri,
            ),
        )


def _parse_azure_connection_string(
    connection_str: str,
) -> dict[Literal["azure_storage_account_name", "azure_storage_account_key"], str]:
    parameters = dict(keyvalue.split("=", maxsplit=1) for keyvalue in connection_str.split(";"))
    return {
        "azure_storage_account_name": parameters.get("AccountName"),
        "azure_storage_account_key": parameters.get("AccountKey"),
    }



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import CliConfig, Dict
from unstructured_ingest.connector.delta_table import DeltaTableWriteConfig, SimpleDeltaTableConfig

CMD_NAME = "delta-table"


@dataclass
class DeltaTableCliConfig(SimpleDeltaTableConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--table-uri"],
                required=True,
                help="the path of the DeltaTable",
            ),
            click.Option(
                ["--version"],
                default=None,
                type=int,
                help="version of the DeltaTable",
            ),
            click.Option(
                ["--storage_options"],
                required=False,
                type=Dict(),
                default=None,
                help="a dictionary of the options to use for the storage backend, "
                "passed in as a json string",
            ),
            click.Option(
                ["--without-files"],
                is_flag=True,
                default=False,
                help="If set, will load table without tracking files.",
            ),
        ]
        return options


@dataclass
class DeltaTableCliWriteConfig(DeltaTableWriteConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--overwrite-schema"],
                is_flag=True,
                default=False,
                help="Flag to overwrite schema of destination table",
            ),
            click.Option(
                ["--drop-empty-cols"],
                is_flag=True,
                default=False,
                help="Flag to drop any columns that have no content",
            ),
            click.Option(
                ["--mode"],
                default="error",
                type=click.Choice(["error", "append", "overwrite", "ignore"]),
                help="How to handle existing data. Default is to error if table already exists. "
                "If 'append', will add new data. "
                "If 'overwrite', will replace table with new data. "
                "If 'ignore', will not write anything if table already exists.",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name=CMD_NAME,
        cli_config=DeltaTableCliConfig,
    )
    return cmd_cls


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name=CMD_NAME,
        cli_config=DeltaTableCliConfig,
        additional_cli_options=[DeltaTableCliWriteConfig],
        write_config=DeltaTableWriteConfig,
    )
    return cmd_cls



import math
import typing as t
from collections import abc
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from pathlib import Path

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSessionHandle,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    ConfigSessionHandleMixin,
    IngestDocCleanupMixin,
    IngestDocSessionHandleMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from atlassian import Jira


@dataclass
class JiraSessionHandle(BaseSessionHandle):
    service: "Jira"


@requires_dependencies(["atlassian"], extras="jira")
def create_jira_object(url, user_email, api_token):
    """
    Creates a jira object for interacting with Jira Cloud.
    Args:
        url: URL to Jira Cloud organization
        user_email: Email for the user with the permissions
        api_token: API Token, generated for the user

    Returns:
        Jira object
    """
    from atlassian import Jira

    jira = Jira(
        url,
        username=user_email,
        password=api_token,
    )

    response = jira.get_permissions("BROWSE_PROJECTS")
    permitted = response["permissions"]["BROWSE_PROJECTS"]["havePermission"]

    if permitted:
        return jira

    else:
        raise ValueError(
            """The user with the provided *user_email* and the *api_token*
                         is not permitted to browse projects for the jira organization
                         for the provided *url*. Try checking user_email, api_token,
                         and the url arguments.""",
        )


@dataclass
class JiraAccessConfig(AccessConfig):
    api_token: str = enhanced_field(sensitive=True)


@dataclass
class SimpleJiraConfig(ConfigSessionHandleMixin, BaseConnectorConfig):
    """Connector config where:
    user_email is the email to authenticate into Atlassian (Jira) Cloud,
    api_token is the api token to authenticate into Atlassian (Jira) Cloud,
    url is the URL pointing to the Atlassian (Jira) Cloud instance,
    list_of_projects is a list of project that is aimed to be ingested.

    Check ...
    for more info on the api_token.
    """

    user_email: str
    access_config: JiraAccessConfig
    url: str
    projects: t.Optional[t.List[str]] = None
    boards: t.Optional[t.List[str]] = None
    issues: t.Optional[t.List[str]] = None

    def create_session_handle(
        self,
    ) -> JiraSessionHandle:
        service = create_jira_object(
            url=self.url, user_email=self.user_email, api_token=self.access_config.api_token
        )
        return JiraSessionHandle(service=service)


@dataclass
class JiraFileMeta:
    """Metadata specifying:
    project_id: id for the jira project that the issue locates in, and
    issue_key: key for the issue that is being reached to.
    """

    project_id: str
    board_id: t.Optional[str]
    issue_key: str
    issue_id: str


# An implementation to obtain nested-defaultdict functionality.
# Keys have default values in a recursive manner, allowing
# limitless templates to parse an api response object.
def nested_object_to_field_getter(object):
    if isinstance(object, abc.Mapping):
        new_object = {}
        for k, v in object.items():
            if isinstance(v, abc.Mapping):
                new_object[k] = FieldGetter(nested_object_to_field_getter(v))
            else:
                new_object[k] = v
        return FieldGetter(new_object)
    else:
        return object


class FieldGetter(dict):
    def __getitem__(self, key):
        value = super().__getitem__(key) if key in self else None
        if value is None:
            value = FieldGetter({})
        return value


def form_templated_string(issue, parsed_fields, c_sep="|||", r_sep="\n\n\n"):
    """Forms a template string via parsing the fields from the API response object on the issue
    The template string will be saved to the disk, and then will be processed by partition."""
    return r_sep.join(
        [
            _get_id_fields_for_issue(issue),
            _get_project_fields_for_issue(parsed_fields),
            _get_dropdown_fields_for_issue(parsed_fields),
            _get_subtasks_for_issue(parsed_fields),
            _get_comments_for_issue(parsed_fields),
            _get_text_fields_for_issue(parsed_fields),
        ],
    )


DEFAULT_C_SEP = " " * 5
DEFAULT_R_SEP = "\n"


def _get_id_fields_for_issue(issue, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    id, key = issue["id"], issue["key"]
    return f"IssueID_IssueKey:{id}{c_sep}{key}{r_sep}"


def _get_project_fields_for_issue(issue, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    if "project" in issue:
        return (
            f"""ProjectID_Key:{issue["project"]["key"]}{c_sep}{issue["project"]["name"]}{r_sep}"""
        )
    else:
        return ""


def _get_dropdown_fields_for_issue(issue, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    return f"""
    IssueType:{issue["issuetype"]["name"]}
    {r_sep}
    Status:{issue["status"]["name"]}
    {r_sep}
    Priority:{issue["priority"]}
    {r_sep}
    AssigneeID_Name:{issue["assignee"]["accountId"]}{c_sep}{issue["assignee"]["displayName"]}
    {r_sep}
    ReporterAdr_Name:{issue["reporter"]["emailAddress"]}{c_sep}{issue["reporter"]["displayName"]}
    {r_sep}
    Labels:{c_sep.join(issue["labels"])}
    {r_sep}
    Components:{c_sep.join([component["name"] for component in issue["components"]])}
    {r_sep}
    """


def _get_subtasks_for_issue(issue):
    return ""


def _get_text_fields_for_issue(issue, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    return f"""
    {issue["summary"]}
    {r_sep}
    {issue["description"]}
    {r_sep}
    {c_sep.join([atch["self"] for atch in issue["attachment"]])}
    {r_sep}
    """


def _get_comments_for_issue(issue, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    return c_sep.join(
        [_get_fields_for_comment(comment) for comment in issue["comment"]["comments"]],
    )


def _get_fields_for_comment(comment, c_sep=DEFAULT_C_SEP, r_sep=DEFAULT_R_SEP):
    return f"{comment['author']['displayName']}{c_sep}{comment['body']}{r_sep}"


def scroll_wrapper(func, results_key="results"):
    def wrapper(*args, **kwargs):
        """Wraps a function to obtain scroll functionality.
        Function needs to be able to accept 'start' and 'limit' arguments."""
        if "number_of_items_to_fetch" in kwargs:
            number_of_items_to_fetch = kwargs["number_of_items_to_fetch"]
            del kwargs["number_of_items_to_fetch"]
        else:
            number_of_items_to_fetch = 100

        kwargs["limit"] = min(100, number_of_items_to_fetch)
        kwargs["start"] = kwargs.get("start", 0)

        all_results = []
        num_iterations = math.ceil(number_of_items_to_fetch / kwargs["limit"])

        for _ in range(num_iterations):
            response = func(*args, **kwargs)
            if isinstance(response, list):
                all_results += func(*args, **kwargs)
            elif isinstance(response, dict):
                if results_key not in response:
                    raise KeyError(
                        "Response object has no known keys to \
                                   access the results, such as 'results' or 'values'.",
                    )
                all_results += func(*args, **kwargs)[results_key]
            kwargs["start"] += kwargs["limit"]

        return all_results[:number_of_items_to_fetch]

    return wrapper


@dataclass
class JiraIngestDoc(IngestDocSessionHandleMixin, IngestDocCleanupMixin, BaseSingleIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing).

    Current implementation creates a Jira connection object
    to fetch each doc, rather than creating a it for each thread.
    """

    connector_config: SimpleJiraConfig
    file_meta: t.Optional[JiraFileMeta] = None
    registry_name: str = "jira"

    @cached_property
    def record_locator(self):  # Values must be JSON-serializable
        """A dictionary with any data necessary to uniquely identify the document on
        the source system."""
        return {
            "base_url": self.connector_config.url,
            "issue_key": self.file_meta.issue_key,
        }

    @cached_property
    @SourceConnectionNetworkError.wrap
    def issue(self):
        """Gets issue data"""
        jira = self.session_handle.service
        return jira.issue(self.file_meta.issue_key)

    @cached_property
    def parsed_fields(self):
        return nested_object_to_field_getter(self.issue["fields"])

    @property
    def grouping_folder_name(self):
        if self.file_meta.board_id:
            return self.file_meta.board_id
        else:
            return self.file_meta.project_id

    @property
    def filename(self):
        download_file = f"{self.file_meta.issue_id}.txt"

        return (
            Path(self.read_config.download_dir) / self.grouping_folder_name / download_file
        ).resolve()

    @property
    def _output_filename(self):
        """Create output file path."""
        output_file = f"{self.file_meta.issue_id}.json"

        return (
            Path(self.processor_config.output_dir) / self.grouping_folder_name / output_file
        ).resolve()

    @property
    def version(self) -> t.Optional[str]:
        return None

    def update_source_metadata(self, **kwargs) -> None:
        exists = bool(self.issue)
        if not exists:
            self.source_metadata = SourceMetadata(
                exists=exists,
            )
            return

        self.source_metadata = SourceMetadata(
            date_created=datetime.strptime(
                self.parsed_fields["created"],
                "%Y-%m-%dT%H:%M:%S.%f%z",
            ).isoformat(),
            date_modified=datetime.strptime(
                self.parsed_fields["updated"],
                "%Y-%m-%dT%H:%M:%S.%f%z",
            ).isoformat(),
            source_url=f"{self.connector_config.url}/browse/{self.file_meta.issue_key}",
            exists=exists,
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["atlassian"], extras="jira")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        document = form_templated_string(self.issue, self.parsed_fields)
        self.update_source_metadata()
        self.filename.parent.mkdir(parents=True, exist_ok=True)

        with open(self.filename, "w", encoding="utf8") as f:
            f.write(document)


@dataclass
class JiraSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    """Fetches issues from projects in an Atlassian (Jira) Cloud instance."""

    connector_config: SimpleJiraConfig
    _jira: t.Optional["Jira"] = field(init=False, default=None)

    @property
    def jira(self) -> "Jira":
        if self._jira is None:
            try:
                self._jira = self.connector_config.create_session_handle().service
            except Exception as e:
                logger.error(f"failed to validate connection: {e}", exc_info=True)
                raise SourceConnectionError(f"failed to validate connection: {e}")
        return self._jira

    @requires_dependencies(["atlassian"], extras="jira")
    def initialize(self):
        _ = self.jira

    def check_connection(self):
        _ = self.jira

    @requires_dependencies(["atlassian"], extras="jira")
    def _get_all_project_ids(self):
        """Fetches ids for all projects in a Jira domain."""
        project_ids = [project["key"] for project in self.jira.projects()]
        return project_ids

    @requires_dependencies(["atlassian"], extras="jira")
    def _get_issues_within_one_project(
        self,
        project_id: str,
    ):
        get_issues_with_scroll = scroll_wrapper(self.jira.get_all_project_issues)
        results = get_issues_with_scroll(project=project_id, fields=["key"])

        return [(issue["key"], issue["id"], None) for issue in results]

    @requires_dependencies(["atlassian"], extras="jira")
    def _get_issue_keys_within_projects(self, project_ids: t.Optional[t.List[str]] = None):
        if project_ids is None:
            # for when a component list is provided, without any projects
            if bool(self.connector_config.boards or self.connector_config.issues):
                return []
            # for when no components are provided. all projects will be ingested
            else:
                return self._get_all_project_ids()

        # for when a component list is provided, including some projects
        issue_keys_all = [self._get_issues_within_one_project(project_id=id) for id in project_ids]

        issue_keys_flattened = [
            (issue_key, issue_id, None)
            for issue_keys_project in issue_keys_all
            for issue_key, issue_id, board_id in issue_keys_project
        ]

        return issue_keys_flattened

    def _get_issues_within_one_board(self, board_id: str):
        get_issues_with_scroll = scroll_wrapper(
            self.jira.get_issues_for_board,
            results_key="issues",
        )
        results = get_issues_with_scroll(board_id=board_id, fields=["key"], jql=None)

        return [(issue["key"], issue["id"], board_id) for issue in results]

    def _get_issue_keys_within_boards(self, board_ids):
        if board_ids is None:
            return []

        issue_keys_all = [self._get_issues_within_one_board(board_id=id) for id in board_ids]

        issue_keys_flattened = [
            (issue_key, issue_id, board_id)
            for issue_keys_board in issue_keys_all
            for issue_key, issue_id, board_id in issue_keys_board
        ]
        return issue_keys_flattened

    def get_issues_info(self, issues):
        issues_info = [self.jira.get_issue(issue, ["key", "id"]) for issue in issues]
        return [(info["key"], info["id"], None) for info in issues_info]

    def get_issue_keys_for_given_components(self):
        issues = []

        if self.connector_config.projects:
            issues += self._get_issue_keys_within_projects(self.connector_config.projects)
        if self.connector_config.boards:
            issues += self._get_issue_keys_within_boards(self.connector_config.boards)
        if self.connector_config.issues:
            issues += self.get_issues_info(self.connector_config.issues)

        return issues

    def get_ingest_docs(self):
        """Fetches all issues in a project."""
        if bool(
            self.connector_config.projects
            or self.connector_config.boards
            or self.connector_config.issues,
        ):
            issue_keys_and_ids = self.get_issue_keys_for_given_components()
        else:
            # gets all issue ids from all projects
            issue_keys_and_ids = self._get_issue_keys_within_projects()

        return [
            JiraIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                file_meta=JiraFileMeta(
                    issue_id=issue_id,
                    issue_key=issue_key,
                    project_id=issue_key.split("-")[0],
                    board_id=board_id,
                ),
            )
            for issue_key, issue_id, board_id in issue_keys_and_ids
        ]



from unstructured_ingest.cli.base.src import BaseSrcCmd

CMD_NAME = "fsspec"


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(cmd_name=CMD_NAME, is_fsspec=True)
    return cmd_cls


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(cmd_name=CMD_NAME, is_fsspec=True)
    return cmd_cls



import typing as t
from abc import ABC
from dataclasses import dataclass, field

from unstructured_ingest.cli.interfaces import CliConfig
from unstructured_ingest.interfaces import BaseConfig


@dataclass
class BaseCmd(ABC):
    cmd_name: str
    cli_config: t.Optional[t.Type[BaseConfig]] = None
    additional_cli_options: t.List[t.Type[CliConfig]] = field(default_factory=list)
    addition_configs: t.Dict[str, t.Type[BaseConfig]] = field(default_factory=dict)
    is_fsspec: bool = False

    @property
    def cmd_name_key(self):
        return self.cmd_name.replace("-", "_")



from unstructured_ingest.v2.processes.connector_registry import (
    add_destination_entry,
    add_source_entry,
)

from .elasticsearch import CONNECTOR_TYPE as ELASTICSEARCH_CONNECTOR_TYPE
from .elasticsearch import elasticsearch_destination_entry, elasticsearch_source_entry
from .opensearch import CONNECTOR_TYPE as OPENSEARCH_CONNECTOR_TYPE
from .opensearch import opensearch_destination_entry, opensearch_source_entry

add_source_entry(source_type=ELASTICSEARCH_CONNECTOR_TYPE, entry=elasticsearch_source_entry)
add_destination_entry(
    destination_type=ELASTICSEARCH_CONNECTOR_TYPE, entry=elasticsearch_destination_entry
)

add_source_entry(source_type=OPENSEARCH_CONNECTOR_TYPE, entry=opensearch_source_entry)
add_destination_entry(
    destination_type=OPENSEARCH_CONNECTOR_TYPE, entry=opensearch_destination_entry
)



# https://developers.notion.com/reference/property-object#number
from dataclasses import dataclass
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.connector.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class NumberProp(FromJSONMixin):
    format: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class Number(DBPropertyBase):
    id: str
    name: str
    number: NumberProp
    type: str = "number"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(number=NumberProp.from_dict(data.pop("number")), **data)


@dataclass
class NumberCell(DBCellBase):
    id: str
    number: Optional[int] = None
    type: str = "number"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        if number := self.number:
            return Div([], str(number))
        return None



import typing as t
from dataclasses import dataclass
from datetime import datetime
from urllib.parse import urlparse

from unstructured_ingest.connector.git import (
    GitIngestDoc,
    GitSourceConnector,
    SimpleGitConfig,
)
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import SourceMetadata
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from github.Repository import Repository


@dataclass
class SimpleGitHubConfig(SimpleGitConfig):
    def __post_init__(self):
        parsed_gh_url = urlparse(self.url)
        path_fragments = [fragment for fragment in parsed_gh_url.path.split("/") if fragment]

        # If a scheme and netloc are provided, ensure they are correct
        # Additionally, ensure that the path contains two fragments
        if (
            (parsed_gh_url.scheme and parsed_gh_url.scheme != "https")
            or (parsed_gh_url.netloc and parsed_gh_url.netloc != "github.com")
            or len(path_fragments) != 2
        ):
            raise ValueError(
                'Please provide a valid URL, e.g. "https://github.com/Unstructured-IO/unstructured"'
                ' or a repository owner/name pair, e.g. "Unstructured-IO/unstructured".',
            )

        # If there's no issues, store the core repository info
        self.repo_path = parsed_gh_url.path

    @SourceConnectionError.wrap
    @requires_dependencies(["github"], extras="github")
    def get_repo(self) -> "Repository":
        from github import Github

        github = Github(self.access_config.access_token)
        return github.get_repo(self.repo_path)


@dataclass
class GitHubIngestDoc(GitIngestDoc):
    connector_config: SimpleGitHubConfig
    registry_name: str = "github"

    @property
    def date_created(self) -> t.Optional[str]:
        return None

    @requires_dependencies(["github"], extras="github")
    def _fetch_file(self):
        from github.GithubException import UnknownObjectException

        try:
            content_file = self.connector_config.get_repo().get_contents(self.path)
        except UnknownObjectException:
            logger.error(f"File doesn't exists {self.connector_config.url}/{self.path}")
            return None

        return content_file

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["requests"], extras="github")
    def _fetch_content(self, content_file):
        import requests

        contents = b""
        if (
            not content_file.content  # type: ignore
            and content_file.encoding == "none"  # type: ignore
            and content_file.size  # type: ignore
        ):
            logger.info("File too large for the GitHub API, using direct download link instead.")
            # NOTE: Maybe add a raise_for_status to catch connection timeout or HTTP Errors?
            response = requests.get(content_file.download_url)  # type: ignore
            if response.status_code != 200:
                logger.info("Direct download link has failed... Skipping this file.")
                return None
            else:
                contents = response.content
        else:
            contents = content_file.decoded_content  # type: ignore
        return contents

    def update_source_metadata(self, **kwargs):
        content_file = kwargs.get("content_file", self._fetch_file())
        if content_file is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return

        date_modified = datetime.strptime(
            content_file.last_modified,
            "%a, %d %b %Y %H:%M:%S %Z",
        ).isoformat()
        self.source_metadata = SourceMetadata(
            date_modified=date_modified,
            version=content_file.etag,
            source_url=content_file.download_url,
            exists=True,
        )

    def _fetch_and_write(self) -> None:
        content_file = self._fetch_file()
        self.update_source_metadata(content_file=content_file)
        contents = self._fetch_content(content_file)
        if contents is None:
            raise ValueError(
                f"Failed to retrieve file from repo "
                f"{self.connector_config.url}/{self.path}. Check logs",
            )
        with open(self.filename, "wb") as f:
            f.write(contents)


@dataclass
class GitHubSourceConnector(GitSourceConnector):
    connector_config: SimpleGitHubConfig

    @requires_dependencies(["github"], extras="github")
    def check_connection(self):
        from github import Consts
        from github.GithubRetry import GithubRetry
        from github.Requester import Requester

        try:
            requester = Requester(
                auth=self.connector_config.access_config.access_token,
                base_url=Consts.DEFAULT_BASE_URL,
                timeout=Consts.DEFAULT_TIMEOUT,
                user_agent=Consts.DEFAULT_USER_AGENT,
                per_page=Consts.DEFAULT_PER_PAGE,
                verify=True,
                retry=GithubRetry(),
                pool_size=None,
            )
            url_base = (
                "/repositories/" if isinstance(self.connector_config.repo_path, int) else "/repos/"
            )
            url = f"{url_base}{self.connector_config.repo_path}"
            headers, _ = requester.requestJsonAndCheck("HEAD", url)
            logger.debug(f"headers from HEAD request: {headers}")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def get_ingest_docs(self):
        repo = self.connector_config.get_repo()
        # Load the Git tree with all files, and then create Ingest docs
        # for all blobs, i.e. all files, ignoring directories
        sha = self.connector_config.branch or repo.default_branch
        git_tree = repo.get_git_tree(sha, recursive=True)
        return [
            GitHubIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                path=element.path,
            )
            for element in git_tree.tree
            if element.type == "blob"
            and self.is_file_type_supported(element.path)
            and (not self.connector_config.file_glob or self.does_path_match_glob(element.path))
        ]



import pytest
from pydantic import Secret, ValidationError

from unstructured_ingest.v2.interfaces import AccessConfig, ConnectionConfig


def test_failing_connection_config():
    class MyAccessConfig(AccessConfig):
        sensitive_value: str

    class MyConnectionConfig(ConnectionConfig):
        access_config: MyAccessConfig

    with pytest.raises(ValidationError):
        MyConnectionConfig(access_config=MyAccessConfig(sensitive_value="this"))


def test_happy_path_connection_config():
    class MyAccessConfig(AccessConfig):
        sensitive_value: str

    class MyConnectionConfig(ConnectionConfig):
        access_config: Secret[MyAccessConfig]

    connection_config = MyConnectionConfig(access_config=MyAccessConfig(sensitive_value="this"))
    assert connection_config



import json
from pathlib import Path
from typing import Optional

from unstructured_ingest.embed.interfaces import AsyncBaseEmbeddingEncoder, BaseEmbeddingEncoder


def validate_embedding_output(original_elements: list[dict], output_elements: list[dict]):
    """
    Make sure the following characteristics are met:
    * The same number of elements are returned
    * For each element that had text, an embeddings entry was added in the output
    * Other than the embedding, nothing about the element was changed
    """
    assert len(original_elements) == len(output_elements)
    for original_element, output_element in zip(original_elements, output_elements):
        if original_element.get("text"):
            assert output_element.get("embeddings", None)
        output_element.pop("embeddings", None)
        assert original_element == output_element


def validate_raw_embedder(
    embedder: BaseEmbeddingEncoder,
    embedder_file: Path,
    expected_dimension: Optional[int] = None,
    expected_is_unit_vector: bool = True,
):
    with open(embedder_file) as f:
        elements = json.load(f)
    all_text = [element["text"] for element in elements]
    single_text = all_text[0]
    dimension = embedder.dimension
    if expected_dimension:
        assert (
            dimension == expected_dimension
        ), f"dimensions {dimension} didn't match expected: {expected_dimension}"
    is_unit_vector = embedder.is_unit_vector
    assert is_unit_vector == expected_is_unit_vector
    single_embedding = embedder.embed_query(query=single_text)
    assert len(single_embedding) == dimension
    embedded_elements = embedder.embed_documents(elements=elements)
    validate_embedding_output(original_elements=elements, output_elements=embedded_elements)


async def validate_raw_embedder_async(
    embedder: AsyncBaseEmbeddingEncoder,
    embedder_file: Path,
    expected_dimension: Optional[int] = None,
    expected_is_unit_vector: bool = True,
):
    with open(embedder_file) as f:
        elements = json.load(f)
    all_text = [element["text"] for element in elements]
    single_text = all_text[0]
    dimension = await embedder.dimension
    if expected_dimension:
        assert (
            dimension == expected_dimension
        ), f"dimension {dimension} didn't match expected: {expected_dimension}"
    is_unit_vector = await embedder.is_unit_vector
    assert is_unit_vector == expected_is_unit_vector
    single_embedding = await embedder.embed_query(query=single_text)
    assert len(single_embedding) == dimension
    embedded_elements = await embedder.embed_documents(elements=elements)
    validate_embedding_output(original_elements=elements, output_elements=embedded_elements)



from pathlib import Path

import pytest
from _pytest.fixtures import TopRequest

from test.integration.connectors.utils.constants import DESTINATION_TAG, SQL_TAG
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from unstructured_ingest.v2.processes.connectors.sql.vastdb import (
    CONNECTOR_TYPE,
    VastdbUploadStager,
    VastdbUploadStagerConfig,
)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_vast_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = VastdbUploadStager(
        upload_stager_config=VastdbUploadStagerConfig(rename_columns_map={"page_number": "page"})
    )
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import json

import pytest

from unstructured_ingest.logger import (
    default_is_data_sensitive,
    hide_sensitive_fields,
    redact_jsons,
)


@pytest.mark.parametrize(
    ("key", "value", "is_sensitive"),
    [
        ("username", "john_smith", False),
        ("password", "13?H%", True),
        ("token", "123", True),
        ("AWS_CREDENTIAL", "aws_credential", True),
        ("AWS_KEY", None, False),
    ],
)
def test_default_is_sensitive(key, value, is_sensitive):
    assert default_is_data_sensitive(key, value) == is_sensitive


def test_hide_sensitive_fields():
    d = {
        "username": "john_smith",
        "password": "13?H%",
        "inner": {
            "token": "123",
            "AWS_KEY": None,
            "inner_j_string": json.dumps(
                {"account_name": "secret name", "client_id": 123, "timestamp": 123}
            ),
        },
    }
    redacted_d = hide_sensitive_fields(d)
    expected_d = {
        "password": "*******",
        "username": "john_smith",
        "inner": {
            "token": "*******",
            "AWS_KEY": None,
            "inner_j_string": json.dumps(
                {"account_name": "*******", "client_id": "*******", "timestamp": 123}
            ),
        },
    }
    assert redacted_d == expected_d


def test_redact_jsons():
    d1 = {
        "username": "john_smith",
        "password": "13?H%",
        "inner": {
            "token": "123",
            "AWS_KEY": None,
            "inner_j_string": json.dumps(
                {"account_name": "secret name", "client_id": 123, "timestamp": 123}
            ),
        },
    }

    d2 = {"username": "tim67", "update_time": 456}
    d3 = {"account_name": "top secret", "host": "http://localhost:8888"}

    sensitive_string = f"Some topic secret info ({json.dumps(d1)} regarding {d2} and {d3})"
    expected_string = (
        'Some topic secret info ({"username": "john_smith", "password": "*******", '
        '"inner": {"token": "*******", "AWS_KEY": null, "inner_j_string": '
        '"{\\"account_name\\": \\"*******\\", \\"client_id\\": \\"*******\\", '
        '\\"timestamp\\": 123}"}} regarding {"username": "tim67", "update_time": 456} '
        'and {"account_name": "*******", "host": "http://localhost:8888"})'
    )
    redacted_string = redact_jsons(sensitive_string)
    assert redacted_string == expected_string



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import (
    CliConfig,
    CliRecursiveConfig,
    DelimitedString,
)
from unstructured_ingest.connector.salesforce import SimpleSalesforceConfig


@dataclass
class SalesforceCliConfig(SimpleSalesforceConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        possible_categories = ["Account", "Case", "Campaign", "EmailMessage", "Lead"]
        options = [
            click.Option(
                ["--username"],
                required=True,
                type=str,
                help="Salesforce username usually looks like an email.",
            ),
            click.Option(
                ["--consumer-key"],
                required=True,
                type=str,
                help="For the Salesforce JWT auth. Found in Consumer Details.",
            ),
            click.Option(
                ["--private-key"],
                required=True,
                type=str,
                help="Path to the private key or its contents for the Salesforce JWT auth. "
                "Key file is usually named server.key.",
            ),
            click.Option(
                ["--categories"],
                default=None,
                required=True,
                type=DelimitedString(choices=possible_categories),
                help="Comma-delimited salesforce categories to download. "
                "Currently only {}.".format(", ".join(possible_categories)),
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name="salesforce",
        cli_config=SalesforceCliConfig,
        additional_cli_options=[CliRecursiveConfig],
    )
    return cmd_cls



import fnmatch
from abc import ABC
from dataclasses import dataclass, field
from typing import Any, Callable, Optional

from pydantic import BaseModel, Field

from unstructured_ingest.v2.interfaces import FileData
from unstructured_ingest.v2.interfaces.process import BaseProcess
from unstructured_ingest.v2.logger import logger


class FiltererConfig(BaseModel):
    file_glob: Optional[list[str]] = Field(
        default=None,
        description="file globs to limit which types of " "files are accepted",
        examples=["*.pdf", "*.html"],
    )
    max_file_size: Optional[int] = Field(
        default=None, description="Max file size to process in bytes"
    )


@dataclass
class Filterer(BaseProcess, ABC):
    config: FiltererConfig = field(default_factory=lambda: FiltererConfig())
    filters: list[Callable[[FileData], bool]] = field(init=False, default_factory=list)

    def __post_init__(self):
        # Populate the filters based on values in config
        if self.config.file_glob is not None:
            self.filters.append(self.glob_filter)
        if self.config.max_file_size:
            self.filters.append(self.file_size_filter)

    def is_async(self) -> bool:
        return False

    def file_size_filter(self, file_data: FileData) -> bool:
        if filesize_bytes := file_data.metadata.filesize_bytes:
            return filesize_bytes <= self.config.max_file_size
        return True

    def glob_filter(self, file_data: FileData) -> bool:
        patterns = self.config.file_glob
        path = file_data.source_identifiers.fullpath
        for pattern in patterns:
            if fnmatch.filter([path], pattern):
                return True
        logger.debug(f"the file {path!r} is discarded as it does not match any given glob.")
        return False

    def run(self, file_data: FileData, **kwargs: Any) -> Optional[FileData]:
        for filter in self.filters:
            if not filter(file_data):
                logger.debug(
                    f"filtered out file data due to {filter.__name__}: {file_data.identifier}"
                )
                return None
        return file_data



# https://developers.notion.com/reference/block#embed
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.attributes import Href
from htmlBuilder.tags import A, Br, Div, HtmlTag

from unstructured_ingest.connector.notion.interfaces import BlockBase
from unstructured_ingest.connector.notion.types.rich_text import RichText


@dataclass
class Embed(BlockBase):
    url: str
    caption: List[RichText] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return False

    @classmethod
    def from_dict(cls, data: dict):
        return cls(caption=[RichText.from_dict(d) for d in data.pop("caption", [])], **data)

    def get_html(self) -> Optional[HtmlTag]:
        texts = []
        if self.url:
            texts.append(A([Href(self.url)], self.url))
        if self.caption:
            texts.append(Div([], [rt.get_html() for rt in self.caption]))
        if not texts:
            return None
        joined = [Br()] * (len(texts) * 2 - 1)
        joined[0::2] = texts

        return Div([], joined)



from __future__ import annotations

import asyncio
import importlib
from functools import wraps
from typing import (
    Callable,
    List,
    Optional,
    TypeVar,
)

from typing_extensions import ParamSpec

_T = TypeVar("_T")
_P = ParamSpec("_P")


def requires_dependencies(
    dependencies: str | list[str],
    extras: Optional[str] = None,
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    """Decorator ensuring required modules are installed.

    Use on functions with local imports to ensure required modules are available and log
    an installation instruction if they're not.

    Args:
        dependencies: Name(s) of module(s) required by the decorated function.
        extras: unstructured-ingest extra which installs required `dependencies`. Defaults to None.

    Raises:
        ImportError: When at least one of the `dependencies` is not available.
    """
    if isinstance(dependencies, str):
        dependencies = [dependencies]

    def decorator(func: Callable[_P, _T]) -> Callable[_P, _T]:
        def run_check():
            missing_deps: List[str] = []
            for dep in dependencies:
                if not dependency_exists(dep):
                    missing_deps.append(dep)
            if len(missing_deps) > 0:
                raise ImportError(
                    f"Following dependencies are missing: {', '.join(missing_deps)}. "
                    + (
                        f"""Please install them using `pip install "unstructured-ingest[{extras}]"`."""  # noqa: E501
                        if extras
                        else f"Please install them using `pip install {' '.join(missing_deps)}`."
                    ),
                )

        @wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs):
            run_check()
            return func(*args, **kwargs)

        @wraps(func)
        async def wrapper_async(*args: _P.args, **kwargs: _P.kwargs):
            run_check()
            return await func(*args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            return wrapper_async
        return wrapper

    return decorator


def dependency_exists(dependency: str):
    try:
        importlib.import_module(dependency)
    except ImportError as e:
        # Check to make sure this isn't some unrelated import error.
        if dependency in repr(e):
            return False
    return True



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseSourceConnector
from unstructured_ingest.logger import logger
from unstructured_ingest.runner.base_runner import Runner
from unstructured_ingest.runner.utils import update_download_dir_remote_url

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.fsspec.s3 import SimpleS3Config


@dataclass
class S3Runner(Runner):
    connector_config: "SimpleS3Config"

    def update_read_config(self):
        self.read_config.download_dir = update_download_dir_remote_url(
            connector_name="s3",
            read_config=self.read_config,
            remote_url=self.connector_config.remote_url,  # type: ignore
            logger=logger,
        )

    def get_source_connector_cls(self) -> t.Type[BaseSourceConnector]:
        from unstructured_ingest.connector.fsspec.s3 import S3SourceConnector

        return S3SourceConnector



import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseDestinationConnector
from unstructured_ingest.runner.writers.base_writer import Writer

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.kafka import KafkaWriteConfig, SimpleKafkaConfig


@dataclass
class KafkaWriter(Writer):
    write_config: "KafkaWriteConfig"
    connector_config: "SimpleKafkaConfig"

    def get_connector_cls(self) -> t.Type[BaseDestinationConnector]:
        from unstructured_ingest.connector.kafka import (
            KafkaDestinationConnector,
        )

        return KafkaDestinationConnector



# https://developers.notion.com/reference/block#synced-block
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import HtmlTag

from unstructured_ingest.connector.notion.interfaces import BlockBase


@dataclass
class OriginalSyncedBlock(BlockBase):
    synced_from: Optional[str] = None
    children: List[dict] = field(default_factory=list)

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls(children=data["children"])

    def get_html(self) -> Optional[HtmlTag]:
        return None


@dataclass
class DuplicateSyncedBlock(BlockBase):
    type: str
    block_id: str

    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return None


class SyncBlock(BlockBase):
    @staticmethod
    def can_have_children() -> bool:
        return True

    @classmethod
    def from_dict(cls, data: dict):
        if "synced_from" in data:
            return OriginalSyncedBlock.from_dict(data)
        else:
            return DuplicateSyncedBlock.from_dict(data)

    def get_html(self) -> Optional[HtmlTag]:
        return None



import json
import os
from dataclasses import dataclass
from pathlib import Path
from uuid import uuid4

import pytest
from _pytest.fixtures import TopRequest
from astrapy import Collection
from astrapy import DataAPIClient as AstraDBClient

from test.integration.connectors.utils.constants import DESTINATION_TAG, SOURCE_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.interfaces import FileData, SourceIdentifiers
from unstructured_ingest.v2.processes.connectors.astradb import (
    CONNECTOR_TYPE,
    AstraDBAccessConfig,
    AstraDBConnectionConfig,
    AstraDBDownloader,
    AstraDBDownloaderConfig,
    AstraDBIndexer,
    AstraDBIndexerConfig,
    AstraDBUploader,
    AstraDBUploaderConfig,
    AstraDBUploadStager,
    DestinationConnectionError,
    SourceConnectionError,
)

EXISTENT_COLLECTION_NAME = "ingest_test_src"
NONEXISTENT_COLLECTION_NAME = "nonexistant"


@pytest.fixture
def connection_config() -> AstraDBConnectionConfig:
    return AstraDBConnectionConfig(
        access_config=AstraDBAccessConfig(
            token=os.environ["ASTRA_DB_APPLICATION_TOKEN"],
            api_endpoint=os.environ["ASTRA_DB_API_ENDPOINT"],
        )
    )


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_APPLICATION_TOKEN", "ASTRA_DB_API_ENDPOINT")
def test_precheck_succeeds_indexer(connection_config: AstraDBConnectionConfig):
    indexer = AstraDBIndexer(
        connection_config=connection_config,
        index_config=AstraDBIndexerConfig(collection_name=EXISTENT_COLLECTION_NAME),
    )
    uploader = AstraDBUploader(
        connection_config=connection_config,
        upload_config=AstraDBUploaderConfig(collection_name=EXISTENT_COLLECTION_NAME),
    )
    indexer.precheck()
    uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_APPLICATION_TOKEN", "ASTRA_DB_API_ENDPOINT")
def test_precheck_succeeds_uploader(connection_config: AstraDBConnectionConfig):
    uploader = AstraDBUploader(
        connection_config=connection_config,
        upload_config=AstraDBUploaderConfig(collection_name=EXISTENT_COLLECTION_NAME),
    )
    uploader.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_APPLICATION_TOKEN", "ASTRA_DB_API_ENDPOINT")
def test_precheck_fails_indexer(connection_config: AstraDBConnectionConfig):
    indexer = AstraDBIndexer(
        connection_config=connection_config,
        index_config=AstraDBIndexerConfig(collection_name=NONEXISTENT_COLLECTION_NAME),
    )
    with pytest.raises(expected_exception=SourceConnectionError):
        indexer.precheck()


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_APPLICATION_TOKEN", "ASTRA_DB_API_ENDPOINT")
def test_precheck_fails_uploader(connection_config: AstraDBConnectionConfig):
    uploader = AstraDBUploader(
        connection_config=connection_config,
        upload_config=AstraDBUploaderConfig(collection_name=NONEXISTENT_COLLECTION_NAME),
    )
    with pytest.raises(expected_exception=DestinationConnectionError):
        uploader.precheck()


@dataclass(frozen=True)
class EnvData:
    api_endpoint: str
    token: str


def get_env_data() -> EnvData:
    return EnvData(
        api_endpoint=os.environ["ASTRA_DB_API_ENDPOINT"],
        token=os.environ["ASTRA_DB_APPLICATION_TOKEN"],
    )


@pytest.fixture
def collection(upload_file: Path) -> Collection:
    random_id = str(uuid4())[:8]
    collection_name = f"utic_test_{random_id}"
    with upload_file.open("r") as upload_fp:
        upload_data = json.load(upload_fp)
    first_content = upload_data[0]
    embeddings = first_content["embeddings"]
    embedding_dimension = len(embeddings)
    my_client = AstraDBClient()
    env_data = get_env_data()
    astra_db = my_client.get_database(
        api_endpoint=env_data.api_endpoint,
        token=env_data.token,
    )
    collection = astra_db.create_collection(collection_name, dimension=embedding_dimension)
    try:
        yield collection
    finally:
        astra_db.drop_collection(collection)


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_API_ENDPOINT", "ASTRA_DB_APPLICATION_TOKEN")
async def test_astra_search_source(
    tmp_path: Path,
):
    env_data = get_env_data()
    collection_name = "ingest_test_src"
    connection_config = AstraDBConnectionConfig(
        access_config=AstraDBAccessConfig(token=env_data.token, api_endpoint=env_data.api_endpoint)
    )
    indexer = AstraDBIndexer(
        index_config=AstraDBIndexerConfig(
            collection_name=collection_name,
        ),
        connection_config=connection_config,
    )
    downloader = AstraDBDownloader(
        connection_config=connection_config,
        download_config=AstraDBDownloaderConfig(download_dir=tmp_path),
    )

    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id=CONNECTOR_TYPE,
            expected_num_files=5,
            expected_number_indexed_file_data=1,
            validate_downloaded_files=True,
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@requires_env("ASTRA_DB_API_ENDPOINT", "ASTRA_DB_APPLICATION_TOKEN")
async def test_astra_search_destination(
    upload_file: Path,
    collection: Collection,
    tmp_path: Path,
):
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="mock file data",
    )
    stager = AstraDBUploadStager()
    env_data = get_env_data()
    uploader = AstraDBUploader(
        connection_config=AstraDBConnectionConfig(
            access_config=AstraDBAccessConfig(
                api_endpoint=env_data.api_endpoint, token=env_data.token
            ),
        ),
        upload_config=AstraDBUploaderConfig(collection_name=collection.name),
    )
    staged_filepath = stager.run(
        elements_filepath=upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=upload_file.name,
    )
    uploader.precheck()
    uploader.run(path=staged_filepath, file_data=file_data)

    # Run validation
    with staged_filepath.open() as f:
        staged_elements = json.load(f)
    expected_count = len(staged_elements)
    current_count = collection.count_documents(filter={}, upper_bound=expected_count * 2)
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from collection: {current_count}"
    )

    # Rerun and make sure the same documents get updated
    uploader.run(path=staged_filepath, file_data=file_data)
    current_count = collection.count_documents(filter={}, upper_bound=expected_count * 2)
    assert current_count == expected_count, (
        f"Expected count ({expected_count}) doesn't match how "
        f"much came back from collection: {current_count}"
    )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_astra_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = AstraDBUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



"""
Box Connector
Box does not make it simple to download files with an App.
First of all, this does not work with a free Box account.
Make sure the App service email is a collaborator for your folder (co-owner or editor)
Make sure you have the 'write all files' application scope
Maybe check 'Make api calls as the as-user header'
REAUTHORIZE app after making any of the above changes
"""

import typing as t
from dataclasses import dataclass

from unstructured_ingest.connector.fsspec.fsspec import (
    FsspecDestinationConnector,
    FsspecIngestDoc,
    FsspecSourceConnector,
    FsspecWriteConfig,
    SimpleFsspecConfig,
)
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import AccessConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies


class AccessTokenError(Exception):
    """There is a problem with the Access Token."""


@dataclass
class BoxWriteConfig(FsspecWriteConfig):
    pass


@dataclass
class BoxAccessConfig(AccessConfig):
    box_app_config: t.Optional[str] = None


@dataclass
class SimpleBoxConfig(SimpleFsspecConfig):
    access_config: BoxAccessConfig = None

    @requires_dependencies(["boxfs"], extras="box")
    def get_access_config(self) -> dict:
        # Return access_kwargs with oauth. The oauth object cannot be stored directly in the config
        # because it is not serializable.
        from boxsdk import JWTAuth

        access_kwargs_with_oauth: dict[str, t.Any] = {
            "oauth": JWTAuth.from_settings_file(
                self.access_config.box_app_config,
            ),
        }
        access_config: dict[str, t.Any] = self.access_config.to_dict()
        access_config.pop("box_app_config", None)
        access_kwargs_with_oauth.update(access_config)

        return access_kwargs_with_oauth


@dataclass
class BoxIngestDoc(FsspecIngestDoc):
    connector_config: SimpleBoxConfig
    registry_name: str = "box"

    @SourceConnectionError.wrap
    @requires_dependencies(["boxfs", "fsspec"], extras="box")
    def get_file(self):
        super().get_file()


@dataclass
class BoxSourceConnector(FsspecSourceConnector):
    connector_config: SimpleBoxConfig

    @requires_dependencies(["boxfs"], extras="box")
    def check_connection(self):
        from boxfs import BoxFileSystem

        try:
            BoxFileSystem(**self.connector_config.get_access_config())
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def __post_init__(self):
        self.ingest_doc_cls: t.Type[BoxIngestDoc] = BoxIngestDoc


@dataclass
class BoxDestinationConnector(FsspecDestinationConnector):
    connector_config: SimpleBoxConfig
    write_config: BoxWriteConfig

    @requires_dependencies(["boxfs", "fsspec"], extras="box")
    def initialize(self):
        super().initialize()

    @requires_dependencies(["boxfs"], extras="box")
    def check_connection(self):
        from boxfs import BoxFileSystem

        try:
            BoxFileSystem(**self.connector_config.get_access_config())
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")



import json
from datetime import datetime
from pathlib import Path
from typing import Callable


def json_serial(obj):
    if isinstance(obj, Path):
        return obj.as_posix()
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def sterilize_dict(data: dict, default: Callable = json_serial) -> dict:
    data_s = json.dumps(data, default=default)
    return json.loads(data_s)



import copy
import typing as t
from dataclasses import dataclass, field
from pathlib import Path

from unstructured_ingest.__version__ import __version__ as unstructured_version
from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.enhanced_dataclass.core import _asdict
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError, WriteError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseDestinationConnector,
    BaseIngestDocBatch,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from pymongo import MongoClient


SERVER_API_VERSION = "1"


def parse_userinfo(userinfo: str) -> t.Tuple[str, str]:
    user, _, passwd = userinfo.partition(":")
    return user, passwd


@dataclass
class MongoDBAccessConfig(AccessConfig):
    uri: t.Optional[str] = enhanced_field(sensitive=True, default=None)


@dataclass
class SimpleMongoDBConfig(BaseConnectorConfig):
    access_config: MongoDBAccessConfig
    host: t.Optional[str] = None
    database: t.Optional[str] = None
    collection: t.Optional[str] = None
    port: int = 27017
    batch_size: int = 100

    @requires_dependencies(["pymongo"], extras="mongodb")
    def generate_client(self) -> "MongoClient":
        from pymongo import MongoClient
        from pymongo.driver_info import DriverInfo
        from pymongo.server_api import ServerApi

        if self.access_config.uri:
            return MongoClient(
                self.access_config.uri,
                server_api=ServerApi(version=SERVER_API_VERSION),
                driver=DriverInfo(name="unstructured", version=unstructured_version),
            )
        else:
            return MongoClient(
                host=self.host,
                port=self.port,
                server_api=ServerApi(version=SERVER_API_VERSION),
            )

    def get_collection(self, client):
        database = client[self.database]
        return database.get_collection(name=self.collection)


@dataclass
class MongoDBDocumentMeta:
    collection: str
    document_id: str
    date_created: str


@dataclass
class MongoDBIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    connector_config: SimpleMongoDBConfig
    document_meta: MongoDBDocumentMeta
    document: dict = field(default_factory=dict)
    registry_name: str = "mongodb"

    @property
    def filename(self):
        return (
            Path(self.read_config.download_dir)
            / self.connector_config.collection
            / f"{self.document_meta.document_id}.txt"
        ).resolve()

    @property
    def _output_filename(self):
        return (
            Path(self.processor_config.output_dir)
            / self.connector_config.collection
            / f"{self.document_meta.document_id}.json"
        )

    def update_source_metadata(self, **kwargs):
        if self.document is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        self.source_metadata = SourceMetadata(
            date_created=self.document_meta.date_created,
            exists=True,
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["pymongo"], extras="mongodb")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        pass

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "host": self.connector_config.host,
            "collection": self.connector_config.collection,
            "document_id": self.document_meta.document_id,
        }


@dataclass
class MongoDBIngestDocBatch(BaseIngestDocBatch):
    connector_config: SimpleMongoDBConfig
    ingest_docs: t.List[MongoDBIngestDoc] = field(default_factory=list)
    list_of_ids: t.List[str] = field(default_factory=list)
    registry_name: str = "mongodb_batch"

    @property
    def unique_id(self) -> str:
        return ",".join(sorted(self.list_of_ids))

    @requires_dependencies(["pymongo"], extras="mongodb")
    def _get_docs(self) -> t.List[dict]:
        """Fetches all documents in a collection."""
        from bson.objectid import ObjectId

        # Note for future. Maybe this could use other client
        client = self.connector_config.generate_client()
        collection = self.connector_config.get_collection(client)
        # MondoDB expects a list of ObjectIds
        list_of_object_ids = []
        for x in self.list_of_ids:
            list_of_object_ids.append(ObjectId(x))
        return list(collection.find({"_id": {"$in": list_of_object_ids}}))

    def get_files(self):
        documents = self._get_docs()
        for doc in documents:
            ingest_doc = MongoDBIngestDoc(
                processor_config=self.processor_config,
                read_config=self.read_config,
                connector_config=self.connector_config,
                document_meta=MongoDBDocumentMeta(
                    collection=self.connector_config.collection,
                    document_id=str(doc.get("_id")),
                    date_created=doc.get("_id").generation_time.isoformat(),
                ),
                document=doc,
            )
            ingest_doc.update_source_metadata()
            del doc["_id"]
            filename = ingest_doc.filename
            flattened_dict = flatten_dict(dictionary=doc)
            str_values = [str(value) for value in flattened_dict.values()]
            concatenated_values = "\n".join(str_values)

            filename.parent.mkdir(parents=True, exist_ok=True)
            with open(filename, "w", encoding="utf8") as f:
                f.write(concatenated_values)

            self.ingest_docs.append(ingest_doc)


@dataclass
class MongoDBSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    connector_config: SimpleMongoDBConfig
    _client: t.Optional["MongoClient"] = field(init=False, default=None)

    @property
    def client(self) -> "MongoClient":
        if self._client is None:
            self._client = self.connector_config.generate_client()
        return self._client

    def check_connection(self):
        try:
            self.client.admin.command("ping")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def initialize(self):
        _ = self.client

    @requires_dependencies(["pymongo"], extras="mongodb")
    def _get_doc_ids(self) -> t.List[str]:
        """Fetches all document ids in a collection."""
        collection = self.connector_config.get_collection(self.client)
        return [str(x) for x in collection.distinct("_id")]

    def get_ingest_docs(self):
        """Fetches all documents in an index, using ids that are fetched with _get_doc_ids"""
        ids = self._get_doc_ids()
        id_batches = [
            ids[
                i
                * self.connector_config.batch_size : (i + 1)  # noqa
                * self.connector_config.batch_size
            ]
            for i in range(
                (len(ids) + self.connector_config.batch_size - 1)
                // self.connector_config.batch_size
            )
        ]

        return [
            MongoDBIngestDocBatch(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                list_of_ids=batched_ids,
            )
            for batched_ids in id_batches
        ]


@dataclass
class MongoDBDestinationConnector(BaseDestinationConnector):
    connector_config: SimpleMongoDBConfig
    _client: t.Optional["MongoClient"] = field(init=False, default=None)

    def to_dict(self, **kwargs):
        """
        The _client variable in this dataclass breaks deepcopy due to:
        TypeError: cannot pickle '_thread.lock' object
        When serializing, remove it, meaning client data will need to be reinitialized
        when deserialized
        """
        self_cp = copy.copy(self)
        if hasattr(self_cp, "_client"):
            setattr(self_cp, "_client", None)
        return _asdict(self_cp, **kwargs)

    @property
    def client(self) -> "MongoClient":
        if self._client is None:
            self._client = self.connector_config.generate_client()
        return self._client

    @requires_dependencies(["pymongo"], extras="mongodb")
    def check_connection(self):
        try:
            self.client.admin.command("ping")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def initialize(self):
        _ = self.client

    @requires_dependencies(["pymongo"], extras="mongodb")
    def write_dict(self, *args, elements_dict: t.List[t.Dict[str, t.Any]], **kwargs) -> None:
        logger.info(
            f"writing {len(elements_dict)} documents to destination "
            f"database {self.connector_config.database}, "
            f"at collection {self.connector_config.collection}",
        )

        collection = self.connector_config.get_collection(self.client)
        try:
            collection.insert_many(elements_dict)
        except Exception as e:
            logger.error(f"failed to write records: {e}", exc_info=True)
            raise WriteError(f"failed to write records: {e}")



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.v2.processes.connectors.weaviate.weaviate import (
    WeaviateAccessConfig,
    WeaviateConnectionConfig,
    WeaviateUploader,
    WeaviateUploaderConfig,
    WeaviateUploadStager,
    WeaviateUploadStagerConfig,
)

if TYPE_CHECKING:
    from weaviate.auth import AuthCredentials
    from weaviate.client import WeaviateClient

CONNECTOR_TYPE = "weaviate-cloud"


class CloudWeaviateAccessConfig(WeaviateAccessConfig):
    access_token: Optional[str] = Field(
        default=None, description="Used to create the bearer token."
    )
    api_key: Optional[str] = None
    client_secret: Optional[str] = None
    password: Optional[str] = None


class CloudWeaviateConnectionConfig(WeaviateConnectionConfig):
    cluster_url: str = Field(
        description="The WCD cluster URL or hostname to connect to. "
        "Usually in the form: rAnD0mD1g1t5.something.weaviate.cloud"
    )
    username: Optional[str] = None
    anonymous: bool = Field(default=False, description="if set, all auth values will be ignored")
    refresh_token: Optional[str] = Field(
        default=None,
        description="Will tie this value to the bearer token. If not provided, "
        "the authentication will expire once the lifetime of the access token is up.",
    )
    access_config: Secret[CloudWeaviateAccessConfig]

    def model_post_init(self, __context: Any) -> None:
        if self.anonymous:
            return
        access_config = self.access_config.get_secret_value()
        auths = {
            "api_key": access_config.api_key is not None,
            "bearer_token": access_config.access_token is not None,
            "client_secret": access_config.client_secret is not None,
            "client_password": access_config.password is not None and self.username is not None,
        }
        existing_auths = [auth_method for auth_method, flag in auths.items() if flag]

        if len(existing_auths) == 0:
            raise ValueError("No auth values provided and anonymous is False")
        if len(existing_auths) > 1:
            raise ValueError(
                "Multiple auth values provided, only one approach can be used: {}".format(
                    ", ".join(existing_auths)
                )
            )

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_api_key_auth(self) -> Optional["AuthCredentials"]:
        from weaviate.classes.init import Auth

        if api_key := self.access_config.get_secret_value().api_key:
            return Auth.api_key(api_key=api_key)
        return None

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_bearer_token_auth(self) -> Optional["AuthCredentials"]:
        from weaviate.classes.init import Auth

        if access_token := self.access_config.get_secret_value().access_token:
            return Auth.bearer_token(access_token=access_token, refresh_token=self.refresh_token)
        return None

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_client_secret_auth(self) -> Optional["AuthCredentials"]:
        from weaviate.classes.init import Auth

        if client_secret := self.access_config.get_secret_value().client_secret:
            return Auth.client_credentials(client_secret=client_secret)
        return None

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_client_password_auth(self) -> Optional["AuthCredentials"]:
        from weaviate.classes.init import Auth

        if (username := self.username) and (
            password := self.access_config.get_secret_value().password
        ):
            return Auth.client_password(username=username, password=password)
        return None

    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_auth(self) -> "AuthCredentials":
        auths = [
            self.get_api_key_auth(),
            self.get_client_secret_auth(),
            self.get_bearer_token_auth(),
            self.get_client_password_auth(),
        ]
        auths = [auth for auth in auths if auth]
        if len(auths) == 0:
            raise ValueError("No auth values provided and anonymous is False")
        if len(auths) > 1:
            raise ValueError("Multiple auth values provided, only one approach can be used")
        return auths[0]

    @contextmanager
    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_client(self) -> Generator["WeaviateClient", None, None]:
        from weaviate import connect_to_weaviate_cloud
        from weaviate.classes.init import AdditionalConfig

        auth_credentials = None if self.anonymous else self.get_auth()
        with connect_to_weaviate_cloud(
            cluster_url=self.cluster_url,
            auth_credentials=auth_credentials,
            additional_config=AdditionalConfig(timeout=self.get_timeout()),
        ) as weaviate_client:
            yield weaviate_client


class CloudWeaviateUploadStagerConfig(WeaviateUploadStagerConfig):
    pass


@dataclass
class CloudWeaviateUploadStager(WeaviateUploadStager):
    upload_stager_config: CloudWeaviateUploadStagerConfig = field(
        default_factory=lambda: WeaviateUploadStagerConfig()
    )


class CloudWeaviateUploaderConfig(WeaviateUploaderConfig):
    pass


@dataclass
class CloudWeaviateUploader(WeaviateUploader):
    connection_config: CloudWeaviateConnectionConfig = field(
        default_factory=lambda: CloudWeaviateConnectionConfig()
    )
    upload_config: CloudWeaviateUploaderConfig = field(
        default_factory=lambda: CloudWeaviateUploaderConfig()
    )
    connector_type: str = CONNECTOR_TYPE


weaviate_cloud_destination_entry = DestinationRegistryEntry(
    connection_config=CloudWeaviateConnectionConfig,
    uploader=CloudWeaviateUploader,
    uploader_config=CloudWeaviateUploaderConfig,
    upload_stager=CloudWeaviateUploadStager,
    upload_stager_config=CloudWeaviateUploadStagerConfig,
)



import typing as t
from dataclasses import dataclass
from typing import Type

from unstructured_ingest.connector.fsspec.fsspec import (
    FsspecDestinationConnector,
    FsspecIngestDoc,
    FsspecSourceConnector,
    FsspecWriteConfig,
    SimpleFsspecConfig,
)
from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.interfaces import AccessConfig
from unstructured_ingest.utils.dep_check import requires_dependencies


@dataclass
class S3AccessConfig(AccessConfig):
    anon: bool = enhanced_field(default=False, overload_name="anonymous")
    endpoint_url: t.Optional[str] = None
    key: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    secret: t.Optional[str] = enhanced_field(default=None, sensitive=True)
    token: t.Optional[str] = enhanced_field(default=None, sensitive=True)


@dataclass
class S3WriteConfig(FsspecWriteConfig):
    pass


@dataclass
class SimpleS3Config(SimpleFsspecConfig):
    access_config: S3AccessConfig = enhanced_field(default=None)


@dataclass
class S3IngestDoc(FsspecIngestDoc):
    connector_config: SimpleS3Config
    remote_file_path: str
    registry_name: str = "s3"

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def get_file(self):
        super().get_file()


@dataclass
class S3SourceConnector(FsspecSourceConnector):
    connector_config: SimpleS3Config

    def __post_init__(self):
        self.ingest_doc_cls: Type[S3IngestDoc] = S3IngestDoc


@dataclass
class S3DestinationConnector(FsspecDestinationConnector):
    connector_config: SimpleS3Config
    write_config: S3WriteConfig

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    def initialize(self):
        super().initialize()



from .bookmark import Bookmark
from .breadcrumb import Breadcrumb
from .bulleted_list_item import BulletedListItem
from .callout import Callout
from .child_database import ChildDatabase
from .child_page import ChildPage
from .code import Code
from .column_list import Column, ColumnList
from .divider import Divider
from .embed import Embed
from .equation import Equation
from .file import File
from .heading import Heading
from .image import Image
from .link_preview import LinkPreview
from .link_to_page import LinkToPage
from .numbered_list import NumberedListItem
from .paragraph import Paragraph
from .pdf import PDF
from .quote import Quote
from .synced_block import DuplicateSyncedBlock, OriginalSyncedBlock, SyncBlock
from .table import Table, TableRow
from .table_of_contents import TableOfContents
from .template import Template
from .todo import ToDo
from .toggle import Toggle
from .unsupported import Unsupported
from .video import Video

__all__ = [
    "Bookmark",
    "Breadcrumb",
    "BulletedListItem",
    "Callout",
    "ChildDatabase",
    "ChildPage",
    "Code",
    "Column",
    "ColumnList",
    "Divider",
    "Embed",
    "Equation",
    "File",
    "Heading",
    "Image",
    "LinkPreview",
    "LinkToPage",
    "NumberedListItem",
    "Paragraph",
    "PDF",
    "Quote",
    "SyncBlock",
    "OriginalSyncedBlock",
    "DuplicateSyncedBlock",
    "Table",
    "TableRow",
    "TableOfContents",
    "Template",
    "ToDo",
    "Toggle",
    "Unsupported",
    "Video",
]



import math
import typing as t
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

from unstructured_ingest.enhanced_dataclass import enhanced_field
from unstructured_ingest.error import SourceConnectionError, SourceConnectionNetworkError
from unstructured_ingest.interfaces import (
    AccessConfig,
    BaseConnectorConfig,
    BaseSingleIngestDoc,
    BaseSourceConnector,
    IngestDocCleanupMixin,
    SourceConnectorCleanupMixin,
    SourceMetadata,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if t.TYPE_CHECKING:
    from atlassian import Confluence


@dataclass
class ConfluenceAccessConfig(AccessConfig):
    api_token: str = enhanced_field(sensitive=True)


@dataclass
class SimpleConfluenceConfig(BaseConnectorConfig):
    """Connector config where:
    user_email is the email to authenticate into Confluence Cloud,
    api_token is the api token to authenticate into Confluence Cloud,
    and url is the URL pointing to the Confluence Cloud instance.

    Check https://developer.atlassian.com/cloud/confluence/basic-auth-for-rest-apis/
    for more info on the api_token.
    """

    user_email: str
    access_config: ConfluenceAccessConfig
    url: str
    max_num_of_spaces: int = 500
    max_num_of_docs_from_each_space: int = 100
    spaces: t.List[str] = field(default_factory=list)


@dataclass
class ConfluenceDocumentMeta:
    """Metadata specifying:
    id for the confluence space that the document locates in,
    and the id of document that is being reached to.
    """

    space_id: str
    document_id: str


def scroll_wrapper(func):
    def wrapper(*args, **kwargs):
        """Wraps a function to obtain scroll functionality."""
        number_of_items_to_fetch = kwargs["number_of_items_to_fetch"]
        del kwargs["number_of_items_to_fetch"]

        kwargs["limit"] = min(100, number_of_items_to_fetch)
        kwargs["start"] = kwargs.get("start", 0)

        all_results = []
        num_iterations = math.ceil(number_of_items_to_fetch / kwargs["limit"])

        for _ in range(num_iterations):
            response = func(*args, **kwargs)
            if isinstance(response, list):
                all_results += func(*args, **kwargs)
            elif isinstance(response, dict):
                all_results += func(*args, **kwargs)["results"]

            kwargs["start"] += kwargs["limit"]

        return all_results[:number_of_items_to_fetch]

    return wrapper


@dataclass
class ConfluenceIngestDoc(IngestDocCleanupMixin, BaseSingleIngestDoc):
    """Class encapsulating fetching a doc and writing processed results (but not
    doing the processing).

    Current implementation creates a Confluence connection object
    to fetch each doc, rather than creating a it for each thread.
    """

    connector_config: SimpleConfluenceConfig
    document_meta: ConfluenceDocumentMeta
    registry_name: str = "confluence"

    # TODO: remove one of filename or _tmp_download_file, using a wrapper
    @property
    def filename(self):
        if not self.read_config.download_dir:
            return None
        return (
            Path(self.read_config.download_dir)
            / self.document_meta.space_id
            / f"{self.document_meta.document_id}.html"
        ).resolve()

    @property
    def _output_filename(self):
        """Create output file path based on output directory, space id and document id."""
        output_file = f"{self.document_meta.document_id}.json"
        return Path(self.processor_config.output_dir) / self.document_meta.space_id / output_file

    @property
    def record_locator(self) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "url": self.connector_config.url,
            "page_id": self.document_meta.document_id,
        }

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["atlassian"], extras="Confluence")
    def _get_page(self):
        from atlassian import Confluence
        from atlassian.errors import ApiError

        try:
            confluence = Confluence(
                self.connector_config.url,
                username=self.connector_config.user_email,
                password=self.connector_config.access_config.api_token,
            )
            result = confluence.get_page_by_id(
                page_id=self.document_meta.document_id,
                expand="history.lastUpdated,version,body.view",
            )
        except ApiError as e:
            logger.error(e)
            return None
        return result

    def update_source_metadata(self, **kwargs):
        """Fetches file metadata from the current page."""
        page = kwargs.get("page", self._get_page())
        if page is None:
            self.source_metadata = SourceMetadata(
                exists=False,
            )
            return
        document_history = page["history"]
        date_created = datetime.strptime(
            document_history["createdDate"],
            "%Y-%m-%dT%H:%M:%S.%fZ",
        ).isoformat()
        if last_updated := document_history.get("lastUpdated", {}).get("when", ""):
            date_modified = datetime.strptime(
                last_updated,
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).isoformat()
        else:
            date_modified = date_created
        version = page["version"]["number"]
        self.source_metadata = SourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            version=version,
            source_url=page["_links"].get("self", None),
            exists=True,
        )

    @SourceConnectionError.wrap
    @requires_dependencies(["atlassian"], extras="confluence")
    @BaseSingleIngestDoc.skip_if_file_exists
    def get_file(self):
        # TODO: instead of having a separate connection object for each doc,
        # have a separate connection object for each process

        result = self._get_page()
        self.update_source_metadata(page=result)
        if result is None:
            raise ValueError(f"Failed to retrieve page with ID {self.document_meta.document_id}")
        self.document = result["body"]["view"]["value"]
        self.filename.parent.mkdir(parents=True, exist_ok=True)
        with open(self.filename, "w", encoding="utf8") as f:
            f.write(self.document)


@dataclass
class ConfluenceSourceConnector(SourceConnectorCleanupMixin, BaseSourceConnector):
    """Fetches body fields from all documents within all spaces in a Confluence Cloud instance."""

    connector_config: SimpleConfluenceConfig
    _confluence: t.Optional["Confluence"] = field(init=False, default=None)

    @property
    def confluence(self) -> "Confluence":
        from atlassian import Confluence

        if self._confluence is None:
            self._confluence = Confluence(
                url=self.connector_config.url,
                username=self.connector_config.user_email,
                password=self.connector_config.access_config.api_token,
            )
        return self._confluence

    @requires_dependencies(["atlassian", "requests"], extras="Confluence")
    def check_connection(self):
        import requests

        url = "rest/api/space"
        try:
            self.confluence.request(method="HEAD", path=url)
        except requests.HTTPError as http_error:
            logger.error(f"failed to validate connection: {http_error}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {http_error}")

    @requires_dependencies(["atlassian"], extras="Confluence")
    def initialize(self):
        self.list_of_spaces = None
        if self.connector_config.spaces:
            self.list_of_spaces = self.connector_config.spaces
            if self.connector_config.max_num_of_spaces:
                logger.warning(
                    """--confluence-list-of-spaces and --confluence-num-of-spaces cannot
                    be used at the same time. Connector will only fetch the
                    --confluence-list-of-spaces that you've provided.""",
                )

    @requires_dependencies(["atlassian"], extras="Confluence")
    def _get_space_ids(self):
        """Fetches spaces in a confluence domain."""

        get_spaces_with_scroll = scroll_wrapper(self.confluence.get_all_spaces)

        all_results = get_spaces_with_scroll(
            number_of_items_to_fetch=self.connector_config.max_num_of_spaces,
        )

        space_ids = [space["key"] for space in all_results]
        return space_ids

    @requires_dependencies(["atlassian"], extras="Confluence")
    def _get_docs_ids_within_one_space(
        self,
        space_id: str,
        content_type: str = "page",
    ):
        get_pages_with_scroll = scroll_wrapper(self.confluence.get_all_pages_from_space)
        results = get_pages_with_scroll(
            space=space_id,
            number_of_items_to_fetch=self.connector_config.max_num_of_docs_from_each_space,
            content_type=content_type,
        )

        doc_ids = [(space_id, doc["id"]) for doc in results]
        return doc_ids

    @requires_dependencies(["atlassian"], extras="Confluence")
    def _get_doc_ids_within_spaces(self):
        space_ids = self._get_space_ids() if not self.list_of_spaces else self.list_of_spaces

        doc_ids_all = [self._get_docs_ids_within_one_space(space_id=id) for id in space_ids]

        doc_ids_flattened = [
            (space_id, doc_id)
            for doc_ids_space in doc_ids_all
            for space_id, doc_id in doc_ids_space
        ]
        return doc_ids_flattened

    def get_ingest_docs(self):
        """Fetches all documents in a confluence space."""
        doc_ids = self._get_doc_ids_within_spaces()
        return [
            ConfluenceIngestDoc(
                connector_config=self.connector_config,
                processor_config=self.processor_config,
                read_config=self.read_config,
                document_meta=ConfluenceDocumentMeta(space_id, doc_id),
            )
            for space_id, doc_id in doc_ids
        ]
